@@ -1,62 +1,69 @@
 package Apache::FileManager;
 
-
 =head1 NAME
 
-Apache::FileManager - apache mod_perl file manager
+Apache::FileManager - Apache mod_perl File Manager
 
 =head1 SYNOPSIS
 
+ # Install in mod_perl enabled apache conf file
+  <Location /FileManager>
+    SetHandler           perl-script
+    PerlHandler          Apache::FileManager
+  </Location>
 
-- Install in mod_perl enabled apache conf file
-     <Location /FileManager>
-       SetHandler           perl-script
-       PerlHandler          Apache::FileManager
-     </Location>
+  (Then point your browser to http://www.yourwebsite.com/FileManager)
 
-
-- Or call from your own mod_perl script
+ # Or call from your own mod_perl script
   use Apache::FileManager;
   my $obj = Apache::FileManager->new();
   $obj->print();
 
+ # Or create your own custom MyFileManager subclass
 
-- Or create your own custom MyFileManager subclass
-package MyFileManager;
-use strict;
-use Apache::FileManager;
+ package MyFileManager;
+ use strict;
+ use Apache::FileManager;
 
-our @ISA = ('Apache::FileManager');
+ our @ISA = ('Apache::FileManager');
 
-sub handler {
-  my $r = shift;
-  my $obj = __PACKAGE__->new();
-  $r->send_http_header('text/html');
-  print "<HTML><HEAD><TITLE>".$r->server->server_hostname." File Manager</TITLE>
-</HEAD>";
-  $obj->print();
-  print "</HTML>";
-}
+ sub handler {
+   my $r = shift;
+   my $obj = __PACKAGE__->new();
+   $r->send_http_header('text/html');
+   print ("
+    <HTML>
+      <HEAD>
+      <TITLE>".$r->server->server_hostname." File Manager</TITLE>
+      </HEAD>
+   ");
+   $obj->print();
+   print "</HTML>";
+ }
 
-.. overload the methods ..
+ # .. overload the methods ..
 
 
 =head1 DESCRIPTION
 
-The Apache::FileManager module is a simple HTML file manager. It provides file manipulations such as cut, copy, paste, delete, rename, extract archive, create directory, and upload files. The interface is clean and simple, and configuration is a breeze.
+The Apache::FileManager module is a simple HTML file manager.  It provides 
+file manipulations such as cut, copy, paste, delete, rename, extract archive, 
+create directory, create file, edit file, and upload files.
 
-For those of you who are up to the challenge, you can configure Apache::FileManager on run on a development server and update your live server htdocs tree with the click on a button. 
+Apache::FileManager also has the ability to rsync the server htdocs tree to 
+another server. With the click of a button.
 
 =head1 PREREQUISITES 
 
-  The following (non-core) perl modules must be installed before installing Apache::FileManager.
+The following (non-core) perl modules must be installed before installing 
+Apache::FileManager.
 
-      Apache::Request => 1.00
-      Apache::File    => 1.01
-      File::NCopy     => 0.32
-      File::Remove    => 0.20
-      Archive::Any    => 0.03
-      CGI::Cookie     => 1.20
+ Apache::Request => 1.00
+ Apache::File    => 1.01
+ File::NCopy     => 0.32
+ File::Remove    => 0.20
+ Archive::Any    => 0.03
+ CGI::Cookie     => 1.20
 
 =head1 SPECIAL NOTES
 
@@ -67,112 +74,127 @@ run the following commands before you begin.
 chown -R nobody /web/xyz/htdocs
 chmod -R 755 /web/xyz/htdocs
 
-The extract functionality only works with tarballs and zips.
+The extract functionality only works with *.tar.gz and *.zip files.
 
 =head1 RSYNC FEATURE
 
-To use the rync functionality you must have ssh, rsync, and the File::Rsync perl module installed on the development server. You also must have an sshd running on the production server.
+To use the rync functionality you must have ssh, rsync, and the File::Rsync 
+perl module installed on the development server. You also must have an sshd 
+running on the production server.
 
-Make sure you always fully qualify your server names so you don't have different values in your known hosts file.
-for example:
-ssh my-machine                -  wrong
-ssh my-machine.subnet.com     -  right
-
-Note: if the ip address of the production_server changes you will need a new known_hosts file.
+Make sure you always fully qualify your server names so you don't have 
+different values in your known hosts file.
 
+ For Example:
+ ssh my-machine                -  wrong
+ ssh my-machine.subnet.com     -  right
 
+Note: If the ip address of the production_server changes you will need to
+create a new known_hosts file.
 
 To get the rsync feature to work do the following:
 
-  #1 log onto the production server
+ #1 log onto the production server
 
-  #2 become root
+ #2 become root
 
-  #3 give web server user (typically nobody) a home area
-     I made mine /usr/local/apache/nobody
-     - production_server> mkdir /usr/local/apache/nobody
-     - edit passwd file and set new home area for nobody
-     - production_server> mkdir /usr/local/apache/nobody/.ssh
+ #3 give web server user (typically nobody) a home area
+   I made mine /usr/local/apache/nobody
+   - production_server> mkdir /usr/local/apache/nobody
+   - edit passwd file and set new home area for nobody
+   - production_server> mkdir /usr/local/apache/nobody/.ssh
 
-  #4 log onto the development server
+ #4 log onto the development server
 
-  #5 become root
+ #5 become root
 
-  #6 give web server user (typically nobody) a home area
-     - dev_server> mkdir /usr/local/apache/nobody
-     - dev_server> chown -R nobody.nobody /usr/local/apache/nobody
-     - edit passwd file and set new home area for nobody
-     - dev_server> su - nobody
-     - dev_server> ssh-keygen -t dsa      (don't use passphrase)
-     - dev_server> ssh production_server (will fail but will make known_hosts file)
-     - log out from user nobody back to root user
-     - dev_server> cd /usr/local/apache/nobody/.ssh
-     - dev_server> scp id_dsa.pub production_server:/usr/local/apache/nobody/.ssh/authorized_keys
-     - dev_server> chown -R nobody.nobody /usr/local/apache/nobody
-     - dev_server> chmod -R 700 /usr/local/apache/nobody
+ #6 give web server user (typically nobody) a home area
+   - dev_server> mkdir /usr/local/apache/nobody
+   - dev_server> chown -R nobody.nobody /usr/local/apache/nobody
+   - edit passwd file and set new home area for nobody
+   - dev_server> su - nobody
+   - dev_server> ssh-keygen -t dsa      (don't use passphrase)
+   - dev_server> ssh production_server 
+     (will fail but will make known_hosts file)
+   - log out from user nobody back to root user
+   - dev_server> cd /usr/local/apache/nobody/.ssh
+   - dev_server> scp id_dsa.pub production_server:/usr/local/apache/nobody/.ssh/authorized_keys
+   - dev_server> chown -R nobody.nobody /usr/local/apache/nobody
+   - dev_server> chmod -R 700 /usr/local/apache/nobody
 
-  #7 log back into the production server
+ #7 log back into the production server
 
-  #8 become root
+ #8 become root
 
-  #9 Do the following commands:
-     - production_server> chown -R nobody.nobody /usr/local/apache/nobody
-     - production_server> chmod -R 700 /usr/local/apache/nobody
+ #9 Do the following commands:
+   - production_server> chown -R nobody.nobody /usr/local/apache/nobody
+   - production_server> chmod -R 700 /usr/local/apache/nobody
 
-You also need to specify the production server in the development server's web conf file. So your conf file should look like this:
+You also need to specify the production server in the development server's 
+web conf file. So your conf file should look like this:
 
-     <Location /FileManager>
-       SetHandler           perl-script
-       PerlHandler          Apache::FileManager
-       PerlSetVar           RSYNC_TO   production_server:/web/xyz
-     </Location>
+ <Location /FileManager>
+   SetHandler           perl-script
+   PerlHandler          Apache::FileManager
+   PerlSetVar           RSYNC_TO   production_server:/web/xyz
+ </Location>
 
-If your ssh path is not /usr/bin/ssh or /usr/local/bin/ssh, you also need to specify the path in the conf file or in the contructor with the directive SSH_PATH.
+If your ssh path is not /usr/bin/ssh or /usr/local/bin/ssh, you also need to 
+specify the path in the conf file or in the contructor with the directive 
+SSH_PATH.
 
 You can also specify RSYNC_TO in the constructor:
-my $obj = Apache::FileManager->new({ RSYNC_TO => "production_server:/web/xyz" });
 
-Also make sure /web/xyz and all files in the tree are readable, writeable, and executable by nobody on both the production server AND the development server.
+  my $obj = Apache::FileManager->new({ 
+    RSYNC_TO => "production_server:/web/xyz" 
+  });
 
+Also make sure /web/xyz and all files in the tree are readable, writeable, and 
+executable by nobody on both the production server AND the development server.
 
 =head1 USING DIFFERENT DOCUMENT ROOT
 
-You can specify a different document root as long as the new document root falls inside of the orginal document root. For example if the document root of a web server is /web/project/htdocs, you could assign the document root to also be /web/project/htdocs/newroot. The directory `newroot` must exist.
+You can specify a different document root as long as the new document root 
+falls inside of the orginal document root. For example if the document root 
+of a web server is /web/project/htdocs, you could assign the document root to 
+also be /web/project/htdocs/newroot. The directory `newroot` must exist.
 
-- Specify different document root in apache conf file
-     <Location /FileManager>
-       SetHandler           perl-script
-       PerlHandler          Apache::FileManager
-       PerlSetVar           DOCUMENT_ROOT /web/project/htdocs/newroot
-     </Location>
+ # Specify different document root in apache conf file
+   <Location /FileManager>
+     SetHandler           perl-script
+     PerlHandler          Apache::FileManager
+     PerlSetVar           DOCUMENT_ROOT /web/project/htdocs/newroot
+   </Location>
 
-- Or specify different document root in your own mod_perl script
-  use Apache::FileManager;
-  my $obj = Apache::FileManager->new({ DOCUMENT_ROOT => '/web/project/htdocs/newroot' });
-  $obj->print();
+ # Or specify different document root in your own mod_perl script
+   use Apache::FileManager;
+   my $obj = Apache::FileManager->new({ 
+     DOCUMENT_ROOT => '/web/project/htdocs/newroot' 
+   });
+   $obj->print();
 
 =head1 SUBCLASSING Apache::FileManager 
 
-Create a new file with the following code:
+ # Create a new file with the following code:
 
-package MyProject::MyFileManager;
-use strict;
-use Apache::FileManager;
-our @ISA = ('Apache::FileManager');
+ package MyProject::MyFileManager;
+ use strict;
+ use Apache::FileManager;
+ our @ISA = ('Apache::FileManager');
 
-  #Add your own methods here
+ #Add your own methods here
 
 1;
 
-The best way to subclass the filemanager would be to copy the methods you want to overload from the Apache::FileManager file to your new subclass. Then change the methods to your liking. If you think you have a great extension, or you feel like you did something better, email me your subclass and I'll merge the relevant parts to the main code.
-
-=head1 TO DO
-
-I need to write documentation on the different methods. Maybe someone else wants to do this.
+The best way to subclass the filemanager would be to copy the methods you want 
+to overload from the Apache::FileManager file to your new subclass. Then change
+the methods to your liking.
 
 =head1 BUGS
 
-There is a bug in File::NCopy that occurs when trying to paste an empty directory. The directory is copied but reports back as 0 directories pasted. The author is in the process of fixing the problem.
+There is a bug in File::NCopy that occurs when trying to paste an empty 
+directory. The directory is copied but reports back as 0 directories pasted. 
+The author is in the process of fixing the problem.
 
 =head1 AUTHOR
 
@@ -183,7 +205,9 @@ E<lt>pmc@cpan.orgE<gt>.
 
 use strict;
 #use warnings;
+use IO::File;
 use Apache::Request;
+use Apache::Util qw(escape_html);
 use Apache::File;
 use File::NCopy  qw(copy);
 use File::Copy   qw(move);
@@ -197,7 +221,7 @@ use Apache::Constants ':common';
 
 require 5.005_62;
 
-our $VERSION = '0.17';
+our $VERSION = '0.19';
 
 sub r  { return Apache::Request->instance( Apache->request ); }
 
@@ -221,9 +245,9 @@ sub handler {
   my $package = __PACKAGE__;
   my $obj = $package->new();
   r->send_http_header('text/html');
-  print "<HTML><HEAD><TITLE>".r->server->server_hostname." File Manager $VERSION</TITLE></HEAD>";
+  r->print("<HTML><HEAD><TITLE>".r->server->server_hostname." File Manager $VERSION</TITLE></HEAD>");
   $obj->print();
-  print "</HTML>";
+  r->print("</HTML>");
   return OK;
 }
 
@@ -232,6 +256,7 @@ sub handler {
 # ---- Call the view ----------------------------------------------
 sub print {
   my $o = shift;
+
   my $view = "view_".$$o{'view'};
   $o->$view();
 }
@@ -242,6 +267,8 @@ sub intialize {
 
   $$o{MESSAGE} = "";
   $$o{JS} = "";
+  $o->{EDIT_COLS} ||= 75;
+  $o->{EDIT_ROWS} ||= 22;
 
 
   # Is this filemanager rsync capable?
@@ -329,7 +356,7 @@ sub intialize {
 #after upload files - view
 sub view_post_upload {
   my $o = shift;
-  print "<SCRIPT>window.opener.document.FileManager.submit(); window.opener.focus(); window.close();</SCRIPT>";
+  r->print("<SCRIPT>window.opener.document.FileManager.submit(); window.opener.focus(); window.close();</SCRIPT>");
   return undef;
 }
 
@@ -337,7 +364,7 @@ sub view_post_upload {
 #after rsync transacation - view
 sub view_post_rsync {
   my $o = shift;
-  print "<CENTER><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0><TR><TD>$$o{MESSAGE}</TD></TR><TR><FORM><TD ALIGN=RIGHT><INPUT TYPE=BUTTON VALUE='close' onclick=\"window.close();\"></TD></FORM></TR></TABLE></CENTER>";
+  r->print("<CENTER><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0><TR><TD>$$o{MESSAGE}</TD></TR><TR><FORM><TD ALIGN=RIGHT><INPUT TYPE=BUTTON VALUE='close' onclick=\"window.close();\"></TD></FORM></TR></TABLE></CENTER>");
   return undef;
 }
 
@@ -350,7 +377,7 @@ sub view_filemanager {
   my ($location, $up_a_href) = $o->html_location_toolbar();
   $up_a_href = "" if !defined($up_a_href);
 
-  print "
+  r->print("
     <!-- Scripts -->
     ".$o->html_javascript."
 
@@ -375,7 +402,7 @@ sub view_filemanager {
       <!-- Footer -->
       ".$o->html_bottom()."
 
-    </FORM>";
+    </FORM>");
 
   return undef;
 }
@@ -386,7 +413,120 @@ sub view_filemanager {
 
 
 
+sub view_pre_editfile {
+  my $o = shift;
 
+  my $editfile = r->param('FILEMANAGER_editfile');
+  my $base = "http://".r->server->server_hostname."/$editfile";
+  $editfile =~ /([^\/]+)$/;
+  my $filename = $1;
+
+  my $fh;
+  if (-T $filename && -w $filename) {
+    $fh = IO::File->new("< ".$filename); 
+  }
+
+  my $message = "<I><FONT COLOR=#990000>".$$o{MESSAGE}."</FONT></I><BR>" if $$o{MESSAGE};
+
+  if (! $fh) {
+    r->print("
+<HTML>
+<HEAD>
+<BODY>
+<CENTER><BR>$message
+<TABLE BORDER=1 CELLPADDING=10 CELLSPACING=0 BGCOLOR=#606060>
+<TR BGCOLOR=WHITE><TD ALIGN=CENTER>
+  could not open file: <I>$base</I> in text writing mode
+</TD></TR>
+<TR BGCOLOR=#efefef><TD ALIGN=RIGHT>
+  <FORM><INPUT TYPE=BUTTON VALUE=close onclick=\"window.close(); return false;\"></FORM>
+</TD></TR>
+</TABLE>
+<BR>
+</CENTER>
+</BODY>
+</HTML>");
+  }
+
+  else {
+
+    my $data;
+    { local $/=undef; $data = scalar(<$fh>); }
+
+    r->print("
+    <SCRIPT>
+      function show_preview () {
+        var f = window.document.FileManagerEditFile;
+        var w = window.open('', 'FileManagerPreviewEditFile','scrollbars,resizable,menubar,status,toolbar');
+        var d = w.document.open();
+        d.write('<HTML><HEAD><BASE HREF=\"$base\"></HEAD>');
+        d.write(f.FILEMANAGER_filedata.value);
+        d.write('</HTML>');
+        d.close();
+        w.focus();
+      }
+      $$o{JS}
+    </SCRIPT>
+
+    <!-- Styles -->
+    ".$o->html_style_sheet()."
+
+      <FORM NAME=FileManagerEditFile ACTION='".r->uri."' METHOD=POST>
+      ".$o->html_hidden_fields()."
+        <INPUT TYPE=HIDDEN NAME=FILEMANAGER_editfile VALUE=\"".r->param('FILEMANAGER_editfile')."\">
+
+        <!-- Header -->
+<TABLE WIDTH=100% CELLPADDING=0 CELLSPAING=0><TR>
+<TD><FONT COLOR=#3a3a3a><B>$base</B></FONT></TD>
+</TR></TABLE>
+
+        <!-- Special message -->
+        $message
+
+        <TABLE CELLPADDING=2 CELLSPACING=2 BORDER=0 WIDTH=100% BGCOLOR=#606060>
+
+          <!-- Toolbar -->
+          <TR><TD ALIGN=CENTER>
+            <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0 WIDTH=90%><TR ALIGN=CENTER>
+ 
+   	    <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='cancel' onclick=\"window.close(); return false;\"></TD>
+
+   	    <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='preview' onclick=\"window.show_preview(); return false;\"></TD>
+
+            <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='save' onclick=\"var f = window.document.FileManagerEditFile; f.FILEMANAGER_cmd.value = 'savefiledata'; f.submit(); return false;\"></TD>
+
+            </TR></TABLE>
+          </TD></TR>
+
+          <!-- file edit box -->
+          <TR><TD ALIGN=CENTER BGCOLOR=#efefef>
+            <TEXTAREA NAME=FILEMANAGER_filedata COLS=$$o{EDIT_COLS} ROWS=$$o{EDIT_ROWS}>$data</TEXTAREA>
+          </TD></TR>
+
+          <!-- Toolbar -->
+          <TR><TD ALIGN=CENTER>
+            <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0 WIDTH=90%><TR ALIGN=CENTER>
+ 
+   	    <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='cancel' onclick=\"window.close(); return false;\"></TD>
+
+   	    <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='preview' onclick=\"window.show_preview(); return false;\"></TD>
+
+            <TD ALIGN=CENTER><INPUT TYPE=BUTTON VALUE='save' onclick=\"var f = window.document.FileManagerEditFile; f.FILEMANAGER_cmd.value = 'savefiledata'; f.submit(); return false;\"></TD>
+
+            </TR></TABLE>
+          </TD></TR>
+
+        </TABLE>
+          
+        <!-- Footer -->
+        ".$o->html_bottom()."
+
+      </FORM>
+    ");
+  }
+
+  return undef;
+}
 
 
 
@@ -444,6 +584,26 @@ sub html_javascript {
   <!--
   var cookie_name = '$cookie_name';
 
+  function select_all () {
+    var f = window.document.FileManager;
+    var ck_stat;
+    var ar = get_ckbox_array();
+
+    if (ar.length > 0) {
+      if (f.FILEMANAGER_last_select_all.value == '1') {
+        ck_stat = false;
+        f.FILEMANAGER_last_select_all.value = '0';
+      } else {
+        ck_stat = true;
+        f.FILEMANAGER_last_select_all.value = '1';
+      }
+    }
+
+    for (var i=0; i < ar.length; i++) {
+      ar[i].checked = ck_stat;        
+    }
+  }
+
   function display_help () {
     var w=window.open('','help','resizable=yes,scrollbars=yes,width=650,height=650');
     var d = w.document.open();
@@ -521,27 +681,57 @@ sub html_javascript {
     return ar;
   }
 
-  // get the number checked
-  function get_num_checked() {
-    var count = 0;
-    var ar = get_ckbox_array();
+
+  // make array of check box objects that are selected
+  function get_sel_ckbox_array() {
+    var sel_ar = new Array();
+    var sel_files = window.document.FileManager.FILEMANAGER_sel_files;
+    var ar = window.get_ckbox_array(); 
+
     for (var i=0; i < ar.length; i++) {
       if (ar[i].checked == true) {
-        count++;
+        sel_ar.push(ar[i]); 
       }
     }
-    return count;
+
+    return sel_ar;
+  }
+
+
+
+
+  // get the number checked
+  function get_num_checked() {
+    var sel_ar = get_sel_ckbox_array();
+    return sel_ar.length;
+  }
+
+  //function to edit file
+  function edit_file () {
+    var sel_ar = get_sel_ckbox_array();
+
+    //make sure there is 1 and only 1 selected file
+    if (sel_ar.length != 1) {
+      window.alert('Please select ONE file to edit by clicking on a check box with the mouse.');
+    }
+
+    else {
+      var f= window.document.FileManager;
+      var cd = escape(f.FILEMANAGER_curr_dir.value);
+      var editfile = escape(sel_ar[0].value);
+      var w = window.open('".r->uri."?FILEMANAGER_cmd=editfile&FILEMANAGER_curr_dir='+cd+'&FILEMANAGER_editfile='+editfile, 'FileManagerEditFile', 'scrollbars,resizable');
+      sel_ar[0].checked = false;
+      w.focus();  
+    }
   }
 
   // make cookie for checked filenames
   function save_names (type) {
-    var cb = get_ckbox_array();
+    var cb = get_sel_ckbox_array();
     var ac = '';
     for (var i=0; i < cb.length; i++) {
-      if (cb[i].checked == true) {
-        ac = ac + cb[i].value + '|';
-        cb[i].checked = false;
-      }
+      ac = ac + cb[i].value + '|';
+      cb[i].checked = false;
     }
     if (ac == '') {
       window.alert('Please select file(s) by clicking on the check boxes with the mouse.');
@@ -572,8 +762,38 @@ sub html_style_sheet {
   return ("
   <STYLE TYPE='text/css'>
     <!-- 
-      A { text-decoration: none; }
-      A:hover	{ text-decoration: underline; }
+BODY {
+background-color: white;
+font-family: serif;
+margin-top: 0;
+margin-right: 0;
+margin-bottom: 0;
+margin-left: 0;
+padding-top: 0;
+padding-bottom: 0;
+padding-right: 1;
+padding-left: 1;
+border-top: 0;
+border-bottom: 0;
+border-right: 0;
+border-left: 0;
+border-style: 0;
+}
+
+A:link {
+color: #990000;
+text-decoration: none;
+}
+
+A:visited {
+color: #990000;
+text-decoration: none;
+}
+
+A:hover {
+color: #990000;
+text-decoration: underline;
+}
     -->
   </STYLE>")
 }
@@ -586,7 +806,9 @@ sub html_hidden_fields {
   return "
 <INPUT TYPE=HIDDEN NAME=FILEMANAGER_curr_dir VALUE='".r->param('FILEMANAGER_curr_dir')."'>
 <INPUT TYPE=HIDDEN NAME=FILEMANAGER_cmd VALUE=''>
-<INPUT TYPE=HIDDEN NAME=FILEMANAGER_arg VALUE=''>";
+<INPUT TYPE=HIDDEN NAME=FILEMANAGER_arg VALUE=''>
+<INPUT TYPE=HIDDEN NAME=FILEMANAGER_last_select_all VALUE='".r->param('FILEMANAGER_last_select_all')."'>
+";
 }
 
 sub html_location_toolbar {
@@ -622,7 +844,10 @@ sub html_cmd_toolbar {
   my @cmds = (
 
    #Refresh
-   "<A HREF=# onclick=\"var f=window.document.FileManager; f.submit();\"><FONT COLOR=WHITE><B>refresh</B></FONT></A>",
+   "<A HREF=# onclick=\"var f=window.document.FileManager; f.submit(); return false;\"><FONT COLOR=WHITE><B>refresh</B></FONT></A>",
+
+   #Edit
+   "<A HREF=# onclick=\"window.edit_file(); return false;\"><FONT COLOR=WHITE><B>edit</B></FONT></A>",
 
    #Cut
    "<A HREF=# onclick=\"window.save_names('cut'); return false;\"><FONT COLOR=WHITE><B>cut</B></FONT></A>",
@@ -654,13 +879,30 @@ sub html_cmd_toolbar {
 \"><FONT COLOR=WHITE><B>delete</B></FONT></A>",
 
   #Rename
-  "<A HREF=# onclick=\"var f=window.document.FileManager; if (get_num_checked() == 0) { window.alert('Please select a file to rename by clicking on a check box with the mouse.'); } else { f.FILEMANAGER_cmd.value='rename'; var rv=window.prompt('enter new name',''); if ((rv != null)&&(rv != '')) { f.FILEMANAGER_arg.value=rv; f.submit(); } } return false;\"><FONT COLOR=WHITE><B>rename</B></FONT></A>",
+  "<A HREF=# onclick=\"var f=window.document.FileManager; if (get_num_checked() != 1) { window.alert('Please select ONE file to rename by clicking on a check box with the mouse.'); } else { var rv=window.prompt('enter new name',''); if ((rv != null)&&(rv != '')) { f.FILEMANAGER_cmd.value='rename'; f.FILEMANAGER_arg.value=rv; f.submit(); } } return false;\"><FONT COLOR=WHITE><B>rename</B></FONT></A>", 
 
   #Extract
   "<A HREF=# onclick=\"var f=window.document.FileManager; if (get_num_checked() == 0) { window.alert('Please select a file to extract by clicking on a check box with the mouse.'); } else { f.FILEMANAGER_cmd.value='extract'; f.submit(); } return false;\"><FONT COLOR=WHITE><B>extract</B></FONT></A>",
 
+  #New File 
+  "<A HREF=# onclick=\"
+var f=window.document.FileManager; 
+var rv = window.prompt('new file name',''); 
+var cd = f.FILEMANAGER_curr_dir.value;
+if (cd != '') {
+  rv = cd+'/'+rv;
+}
+if ((rv != null)&&(rv != '')) { 
+  var w = window.open('".r->uri."?FILEMANAGER_cmd=editfile&FILEMANAGER_curr_dir='+escape(cd)+'&FILEMANAGER_editfile='+escape(rv), 'FileManagerEditFile', 'scrollbars,resizable'); 
+  w.focus(); 
+} else if (rv == '') {
+  window.alert('can not create blank file names');
+}
+return false;
+\"><FONT COLOR=WHITE><B>new file</B></FONT></A>",
+
   #New Directory
-  "<A HREF=# onclick=\"var f=window.document.FileManager; var rv=window.prompt('new directory name',''); if ((rv != null)&&(rv != '')) { f.FILEMANAGER_arg.value=rv; f.FILEMANAGER_cmd.value='mkdir'; f.submit(); } return false;\"><FONT COLOR=WHITE><B>new directory</B></FONT></A>",
+  "<A HREF=# onclick=\"var f=window.document.FileManager; var rv=window.prompt('new directory name',''); if ((rv != null)&&(rv != '')) { f.FILEMANAGER_arg.value=rv; f.FILEMANAGER_cmd.value='mkdir'; f.submit(); } else if (rv == '') { window.alert('can not create blank directory names'); } return false;\"><FONT COLOR=WHITE><B>new directory</B></FONT></A>",
 
   #Upload
   "<A HREF=# onclick=\"window.print_upload(); return false;\"><FONT COLOR=WHITE><B>upload<B></FONT></A>"
@@ -715,7 +957,7 @@ sub html_file_list {
       $last_modified = "--";
       $size = "<TD ALIGN=CENTER>--</TD>";
       $type = "/"; # "/" designates "directory"
-      $link = "<A HREF=# onclick=\"var f=window.document.FileManager; f.FILEMANAGER_curr_dir.value='$curr_dir"."$file'; f.submit(); return false;\"><FONT COLOR=#006699>$file$type</FONT></A>";
+      $link = "<A HREF=# onclick=\"var f=window.document.FileManager; f.FILEMANAGER_curr_dir.value='".&escape_html($curr_dir.$file)."'; f.submit(); return false;\"><FONT COLOR=#006699>$file$type</FONT></A>";
     }
 
     #must be a file
@@ -756,14 +998,14 @@ sub html_file_list {
       my $href = $curr_dir;
       $href = $fake_doc_root."/".$href if $fake_doc_root;
 
-      $link = "<A HREF=\"/$href"."$file?nossi=1\" TARGET=_blank><FONT COLOR=BLACK>$file$type</FONT></A>";
+      $link = "<A HREF=\"/$href"."$file?nossi=1\" TARGET=_blank><FONT COLOR=BLACK>".&escape_html($file.$type)."</FONT></A>";
     }
 
     $acum .= "
 <TR BGCOLOR=#$bgcolor>
 <TD><INPUT TYPE=CHECKBOX NAME=FILEMANAGER_sel_files VALUE='$curr_dir"."$file'></TD>
 <TD>$link</TD>
-<TD ALIGN=CENTER>$last_modified</TD>
+<TD ALIGN=CENTER NOWRAP>$last_modified</TD>
 $size
 </TR>";
 
@@ -782,7 +1024,7 @@ $size
 
 <!-- Headers -->
 <TR BGCOLOR=#606060>
-<TD WIDTH=1%>&nbsp;</TD>
+<TD WIDTH=1% ALIGN=CENTER><A HREF=# onclick=\"window.select_all(); return false;\"><FONT COLOR=WHITE>+</FONT></A></TD>
 <TD WIDTH=80%><FONT COLOR=WHITE><B>filename</B></FONT></TD>
 <TD WIDTH=15% ALIGN=CENTER NOWRAP><FONT COLOR=WHITE><B>last modified</B></FONT></TD>
 <TD WIDTH=4% ALIGN=CENTER><FONT COLOR=WHITE><B>size</B></FONT></TD>
@@ -875,8 +1117,8 @@ sub filename_esc {
 
 sub formated_date {
   my $o = shift;
-  my $date = shift;
-  return strftime "%D", localtime($date);
+  my $date = shift || time;
+  return strftime "%D %l:%M %P", localtime($date);
 }
 
 sub get_clip_board {
@@ -931,6 +1173,42 @@ sub get_clip_board {
 ###############################################################################
 # -- Commands (called via form input from method execute_cmds or manually) -- #
 ###############################################################################
+
+sub cmd_savefiledata {
+  my $o = shift;
+  my $base = r->param('FILEMANAGER_editfile');
+  $base =~ /([^\/]+)$/;
+  my $filename = $1;
+  remove $filename;
+  my $fh = IO::File->new("> ".$filename);
+  print $fh scalar(r->param('FILEMANAGER_filedata'));
+  $$o{MESSAGE} = "file saved";
+  $$o{view} = "pre_editfile";
+  return undef;
+}
+
+sub cmd_editfile {
+  my $o = shift;
+
+  my $base = r->param('FILEMANAGER_editfile');
+  $base =~ /([^\/]+)$/;
+  my $filename = $1;
+
+  if (! -e $filename) {
+    my $fh = IO::File->new("> ".$filename);
+    if ($fh) {
+      $$o{JS} .= ("
+if (window.opener && window.opener.document.FileManager) {
+  window.opener.document.FileManager.submit();
+}
+");
+    }
+  }
+  $$o{view} = "pre_editfile";
+}
+
+
+
 sub cmd_paste {
   my $o = shift;
   my $arg1 = shift;
